"""quixote.publish

Logic for publishing modules and objects on the Web.
"""

__revision__ = "$Id: publish.py,v 1.79 2001/06/01 21:14:20 nascheme Exp $"

import sys, traceback, os, cStringIO
import time, types, socket, re

from quixote import errors, config
from quixote.util import fcgi
from quixote.zope import HTTPRequest, HTTPResponse

def get_module(name):
    """Get a module object by name."""
    __import__(name)
    module = sys.modules[name]
    return module
    
def _get_module_filename(module):
    filename = module.__file__
    if os.path.isdir( filename ):
        filename = os.path.join( filename, '__init__.py')
    elif filename[ -4: ] == '.pyc': filename = filename[:-1]

    return filename

# Error message to dispay when DISPLAY_EXCEPTIONS in config file is not
# true.  Note that SERVER_ADMIN must be fetched from the environment and
# plugged in here -- we can't do it now because the environment isn't
# really setup for us yet if running as a FastCGI script.
INTERNAL_ERROR_MESSAGE = """\
<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
<html>
<head><title>Internal Server Error</title></head>
<body>
<h1>Internal Server Error</h1>
<p>An internal error occurred while handling your request.</p>

<p>Please contact the server administrator (%s) and inform them of
the time the error occurred, and anything you might have done to
trigger the error.</p>

<p>If you are the server administrator, more information may be
available in either the server's error log or Quixote's error
log.</p>
</body>
</html>
"""

class Publisher:

    def __init__ (self, root_namespace, config=None):
        from quixote.config import Config
        global _publisher

        if _publisher is not None:
            raise RuntimeError, "only one instance of Publisher allowed"
        _publisher = self

        if type(root_namespace) is types.StringType:
            self.root_namespace = get_module(root_namespace)
        else:
            # Should probably check that root_namespace is really a
            # namespace, ie. a module, class, or instance -- but it's
            # tricky to know if something is really a class or instance
            # (because of ExtensionClass), and who knows what other
            # namespaces are lurking out there in the world?
            self.root_namespace = root_namespace

        self.exit_now = 0
        self.access_log = None

        # Initialize default config object with all the default values from
        # the config variables at the top of the config module, ie. if
        # DEBUG_LOG is set to "/var/log/quxiote-debug.log", then
        # config.DEBUG_LOG will also be "/var/log/quixote-debug.log".  If
        # application FCGI/CGI scripts need to override any of these
        # defaults, they can do so by direct manipulation of the config
        # object, or by reading a config file:
        #   app.read_config("myapp.conf")
        if config is None:
            self.config = Config()
            self.config.read_default()
        else:
            self.set_config(config)

    def configure (self, **kwargs):
        self.config.set_from_dict(kwargs)

    def read_config (self, filename):
        self.config.read_file(filename)

    def set_config (self, config):
        from quixote.config import Config
        if not isinstance(config, Config):
            raise TypeError, "'config' must be a Config instance"
        self.config = config

    def install_imphooks (self):
        from quixote import imphooks
        imphooks.install()

    def setup_logs (self):
        """Redirect stdout: either to /dev/null or a debug log file
        (depending on whether DEBUG_LOG is defined in config.py).
        Note that this is not a real redirect in that it doesn't affect
        child processes; this just reassigns 'sys.stdout'.
        """

        if self.config.debug_log is None:
            # If no debug log file is configured, debugging output goes
            # to the bit-bucket.
            sys.stdout = open('/dev/null', 'w')
        else:
            sys.stdout = open(self.config.debug_log, 'a', 1)

        if self.config.access_log is not None:
            try:
                self.access_log = open(self.config.access_log, 'a', 1)
            except IOError, exc:
                sys.stderr.write("error opening access log %s: %s\n" 
                                 % (`self.config.access_log`, exc.strerror))

    def shutdown_logs (self):
        """Restore 'sys.stdout' to what it was before calling
        'setup_logs()'.
        """
        if sys.stdout is sys.__stdout__:
            raise RuntimeError, "'setup_logs()' never called"
        sys.stdout.close()
        sys.stdout = sys.__stdout__

    def debug (self, msg):
        """Print a debugging message to stderr (not stdout, since it will
        disappear if DEBUG_LOG is not set).  What happens to this error message
        depends on your HTTP server; with Apache, it will wind up in Apache's
        error log.  Note that the ERROR_LOG config variable has no effect here,
        since it's only used for application errors.
        """
        # Note that DEBUG_LOG controls where *stdout* goes, but debug messages
        # handled by this function go to *stderr*.  That's because DEBUG_LOG is
        # for debugging Quixote applications, and this function is for
        # debugging Quixote itself.  In other words, if you're hacking on a
        # Quixote app, just set DEBUG_LOG and print your debug messages to
        # stdout.
        from time import time, ctime

        # This is a bogus way to format date/time, but it's what Apache uses in
        # its error_log -- yechh.
        timestamp = ctime(time())
        sys.stderr.write("[%s] [quixote %d] %s\n" % (timestamp, os.getpid(),
                                                     msg))


    def parse_request (self, stdin, env):
        """Parse the request information and return a HTTPRequest object"""
        global _request

        self.start_time = time.time()
        request = HTTPRequest.HTTPRequest(stdin, env)
        _request = request
        request.processInputs()
        return request

    def start_request (self, request):
        """Called at the start of each request.  Overridden by
        SessionPublisher to handle session details."""
        pass

    def log_request (self, request):
        """Log a request in the access_log file."""
        if self.access_log is not None:
            if request.session:
                user = request.session.user
            else:
                user = None
            now = time.time()
            seconds = now - self.start_time
            now = time.strftime('%Y-%m-%d %H:%M:%S',
                                time.localtime(time.time()) )
            self.access_log.write('%s %s %s "%s %s %s" %s %0.2fsec\n' %
                                   (request.environ.get('REMOTE_ADDR'),
                                    str(user),
                                    now,
                                    request.environ.get('REQUEST_METHOD'),
                                    request.environ.get('REQUEST_URI'),
                                    request.environ.get('SERVER_PROTOCOL'),
                                    request.response.status,
                                    seconds
                                   ) )


    def finish_successful_request (self, request):
        """Called at the end of a successful request.  Overridden by
        SessionPublisher to handle session details."""
        pass

    def finish_interrupted_request (self, request, exc):
        """Called at the end of an interrupted request.  Requests are
        interrupted by raising a PublishError exception.  This method should
        return a string object which will be used as the result of the
        request.
        """
        # Get output from the exception; this may make changes to the
        # user's session object.
        request.response = HTTPResponse.HTTPResponse()
        # Test if the exception formats itself
        msg = exc.format(request)
        if msg is not None:
            return msg
        # No, use a standard format
        if isinstance(exc, errors.TraversalError):
            request.response.setStatus(404)
            return self.format_traversal_error(request, exc)
        elif isinstance(exc, errors.QueryError):
            return self.format_query_error(request, exc)
        elif isinstance(exc, errors.AccessError):
            request.response.setStatus(403) # forbidden
            return self.format_access_error(request, exc)
        else:
            raise RuntimeError, "unknown publish exception %s" % `exc`

    def format_query_error (self, request, exc):
        """Format a response after a QueryError exception.  Applications
        may wish to subclass this method."""
        request.response.setHeader("Content-Type", "text/plain")
        return str(exc)

    def format_traversal_error (self, request, exc):
        """Format a response after a TraversalError exception.  Applications
        may wish to subclass this method."""
        from cgi import escape
        page = """\
        <html>
          <head><title>Error: Not Found</title></head>
          <body>%s</body>
        </html>
        """
        msg = 'The requested link "%s" was not found' % exc.path
        if not self.config.secure_errors and exc.message:
            msg += ": %s" % exc.message
        return page % escape(msg)

    def format_access_error (self, request, exc):
        """Format a response after a AccessError exception.  Applications
        may wish to subclass this method."""
        request.response.setHeader("Content-Type", "text/plain")
        return str(exc)

    def format_internal_error (self, request, error_message):
        """Format a response after a non-PublishError exception.
        Applications may wish to subclass this method but they should do
        so carefully.  The application could be broken badly when this
        method is called.  This probably best to keep the formatting
        simple."""
        if self.config.display_exceptions:
            request.response.setHeader("Content-Type", "text/plain")
            return error_message
        else:
            request.response.setHeader("Content-Type", "text/html")
            admin = os.environ.get('SERVER_ADMIN',
                                   "<i>email address unknown</i>")
            return INTERNAL_ERROR_MESSAGE % admin

    def finish_failed_request (self, request):
        """Called at the end of an failed request.  Any exception (other
        than PublishError) cause a request to fail.  This method should
        return a string object which will be used as the result of the
        request.
        """
        # build new response to be safe
        original_response = request.response
        request.response = HTTPResponse.HTTPResponse()
        #self.debug("caught an error (%s), reporting it." %
        #           sys.exc_info()[1])

        exc_type, exc_value, tb = sys.exc_info()
        error_file = cStringIO.StringIO()

        #self.debug("report_error: exception = %s: %s" % (exc_type, exc_value))

        # record the time of the error
        now = time.localtime(time.time())
        error_file.write('Error at time: %s\n' %
                         time.strftime('%Y-%m-%d %H:%M:%S', now) )
        # format the traceback
        error_file.write('Traceback\n'
                         '=========\n')
        traceback.print_exception(exc_type, exc_value, tb, file=error_file)
        # include request and response dumps
        error_file.write('\nRequest\n'
                           '=======\n')
        error_file.write(request.dump())
        error_file.write('\nOriginal Response\n'
                           '=================\n')
        error_file.write(str(original_response))
        error_file.write('='*65 + '\n')

        error_msg = error_file.getvalue()
        if self.config.error_log is None:
            sys.stderr.write(error_msg)
        else:
            try:
                errlog = open(self.config.error_log, 'a')
            except IOError, exc:
                sys.stderr.write("error opening error log %s: %s\n" 
                                 % (`self.config.error_log`, exc.strerror))
            else:
                errlog.write(error_msg)
                if request.session is not None:
                    request.session.dump(file=errlog)
                errlog.close()

        if self.config.error_email:
            self.mail_error(error_msg)

        request.response.setStatus(500)
        return self.format_internal_error(request, error_msg)
        
    def mail_error (self, msg):
        """Send an email notifying someone of a traceback."""

        cmd = "/usr/lib/sendmail -f Quixote '%s'" % self.config.error_email
        pipe = os.popen(cmd, 'w')
        pipe.write('Subject: Quixote Traceback\n')
        pipe.write('From: Quixote on %s <%s>\n' % (socket.gethostname(),
                                                   self.config.error_email))
        pipe.write('To: %s\n' % self.config.error_email)
        pipe.write('\n')
        pipe.write(msg)
        pipe.close()

    def get_component (self, container, component, path, request):
        """Get one component of a path from a namespace."""

        # First security check: if the container doesn't even have an
        # _q_exports list, fail now: all Quixote-traversable namespaces
        # (modules, packages, instances) must have an export list!
        if not hasattr(container, '_q_exports'):
            raise errors.TraversalError(request.get_path(), 
                "%s has no _q_exports list" % repr(container))

        # Second security check: call _q_access function if it's present.
        if hasattr(container, '_q_access'):
            # will raise AccessError if access failed
            container._q_access(request)
            
        # Third security check: make sure the current name component is
        # in the export list or is _q_index.
        if (component == "_q_index") or (component in container._q_exports):
            is_module = type(container) is types.ModuleType
            has_component = hasattr(container, component)
            if is_module and not has_component:
                # get next module
                mod_name = container.__name__ + '.' + component
                # If we get an ImportError here we don't catch it.  It means
                # that either someone exported something that doesn't exist or
                # an exception was raised from deeping in the code.  A
                # traceback should be generated in either case.
                object = get_module(mod_name)

            elif has_component:
                object = getattr(container, component)

            else: 
                # not a module, doesn't have 'component'
                raise errors.TraversalError(request.get_path(),
                    "object %s has no attribute '%s'" % (`container`, component))
        else:
            # Component is not in exports list.  Check if namespace provides a
            # '_q_getname()' function, which translates an arbitrary string
            # into an object that we continue traversing.  This is very handy;
            # it lets you put user-space objects into your URL-space,
            # eliminating the need for digging ID strings out of a query, or
            # checking PATHINFO after Quixote's done with it.  But it is a
            # compromise to security: it opens up the traversal algorithm to
            # arbitrary names not listed in _q_exports!
            if hasattr(container, "_q_getname"):
                object = container._q_getname(request, component)
            else:
                raise errors.TraversalError(request.get_path(),
                    ("%s not in _q_exports list for %s" %
                     (repr(component), container)))
        
        return object

    _slash_pat = re.compile("//*")

    def traverse_url (self, path, request):
        """traverse_url(path : string, request : HTTPRequest) -> object : any

        Perform traversal based on the provided path, starting at the root
        object.  It returns the script name and path info values for
        the arrived-at object, along with the object itself.

        It's expected that the final object is something callable like a
        function or a method; intermediate objects along the way will
        usually be packages or modules.

        To prevent crackers from writing URLs that traverse private
        objects, every package, module, or object along the way must have
        an _q_exports attribute containing a list of publicly visible
        names.  Not having an _q_exports attribute is an error, though
        having _q_exports be an empty list is OK (though useless).  If a
        component of the path isn't in _q_exports, that also produces an
        error.
        """
        
        # replace repeated slashes with a single slash
        if path.find("//") != -1:
            path = self._slash_pat.sub("/", path)

        # split path apart; /foo/bar/baz  -> ['foo', 'bar', 'baz']
        #                   /foo/bar/     -> ['foo', 'bar', '']
        path_components = path[1:].split('/')

        # Traverse starting at the root
        object = self.root_namespace

        # Loop over the components of the path
        for component in path_components:
            if component == "":
                # "/q/foo/" == "/q/foo/_q_index"
                component = "_q_index"
            object = self.get_component(object, component, path, request)

        if not callable(object):
            # We went through all the components of the path and ended up at
            # something which isn't callable, like a module or an instance
            # without a __call__ method.
            if path[-1] != '/':
                if not request.form and self.config.fix_trailing_slash:
                    # This is for the convenience of users who type in paths.
                    # Repair the path and redirect.  This should not happen for
                    # URLs within the site.
                    path = "%s/%s/" % (self.config.url_prefix,
                                       "/".join(path_components))
                    def redirect_slash(request, path=path):
                        return request.redirect(path, permanent=1)
                    return redirect_slash
                else:
                    # Automatic redirects disabled or there is form data.  If
                    # there is form data then the programmer is using the
                    # wrong path.  A redirect won't work if the form data came
                    # from a POST anyhow.
                    raise errors.TraversalError(path, "missing trailing slash?")
            else:
                raise error.TraversalError(path, "uncallable object %s" % `object`)

        # Determine the script name and path info
        script_name = "/".join([self.config.url_prefix] + path_components)
        os.environ['SCRIPT_NAME'] = script_name
        os.environ['PATH_INFO'] = '' # we always traverse the entire path

        return object

    def try_publish (self, request, path):
        """try_publish(request : HTTPRequest, path : string) -> string
        
        The master method that does all the work for a single request.  Uses
        traverse_url() to get a callable object.  The object is called and
        the output is returned.  Exceptions are handled by the caller. 
        """

        self.start_request(request)

        # Traverse package to a (hopefully-) callable object
        object = self.traverse_url(os.environ['PATH_INFO'], request)

        if not callable(object):
            raise RuntimeError, "object not callable: %s" % repr(object)

        # Call object
        try:
            output = object(request)
        except SystemExit:
            output = "SystemExit exception caught, shutting down"
            self.debug(output)
            self.exit_now = 1

        if output is None:
            raise RuntimeError, 'Callable %s returned None' % repr(object)

        # The callable ran OK, commit any changes to the session
        self.finish_successful_request(request)

        return output

    _gzip_header = ("\037\213" # magic
                    "\010" # compression method
                    "\000" # flags
                    "\000\000\000\000" # time, who cares?
                    "\002"
                    "\377")

    def compress_output (self, request, output):
        import zlib, struct

        encoding = request.get_encoding(["gzip", "x-gzip"])
        n = len(output)
        if n > 200 and encoding:
            co = zlib.compressobj(6, zlib.DEFLATED, -zlib.MAX_WBITS,
                                  zlib.DEF_MEM_LEVEL, 0)
            chunks = [self._gzip_header,
                      co.compress(output),
                      co.flush(),
                      struct.pack("<ll", zlib.crc32(output), len(output))]
            output = "".join(chunks)
            #self.debug("gzip (original size %d, ratio %.1f)" % 
            #           (n, float(n)/len(output)))
            request.response.setHeader("Content-Encoding", encoding)
        return output

    def publish (self, stdin, stdout, stderr):
        """publish(stdin : file, stdout : file, stderr : file)

        Process a single request.  The try_publish() method will be called
        to do the work and exceptions will be handled here.
        """

        request = self.parse_request(stdin, os.environ)

        try:
            output = self.try_publish(request, os.environ['PATH_INFO'])

        except errors.PublishError, exc:
            # Exit the publishing loop and return a result right away.
            output = self.finish_interrupted_request(request, exc)

        except:
            # some other exception, generate error messages to the logs, etc.
            output = self.finish_failed_request(request)

        self.log_request(request)

        if self.config.compress_pages:
            output = self.compress_output(request, str(output))

        # Output results from Response object
        request.response.setBody(output)
        stdout.write(str(request.response))

    def publish_cgi (self):
        """publish_cgi() -> None

        Entry point from CGI and FCGI scripts; it will repeatedly do the
        publish() function until there are no more requests (only once for
        CGI scripts).
        """

        while fcgi.isFCGI() and not self.exit_now:
            f = fcgi.FCGI()
            os.environ = f.env
            self.publish(f.inp, f.out, f.err)
            f.Finish()
            if self.config.run_once:
                break


# Publisher singleton, one of these per process
_publisher = None

# The current request being processed.
_request = None

def get_publisher():
    global _publisher
    return _publisher

def get_request():
    global _request
    return _request

def get_session():
    global _request
    return _request.session

def get_user():
    global _request
    session = _request.session
    if session is None:
        return None
    else:
        return session.user

class SessionPublisher (Publisher):
    
    def __init__ (self, root_namespace):
        Publisher.__init__(self, root_namespace)
        self.session_mgr = None

    def set_session_manager (self, session_mgr):
        from quixote.session import set_session_manager
        set_session_manager(session_mgr)
        self.session_mgr = session_mgr

    def start_request (self, request):
        # Get the session object and stick it onto the request
        request.session = self.session_mgr.get_session(request)
        request.session.start_request(request)

    def finish_successful_request (self, request):
        if request.session is not None:
            request.session.finish_request(request)
        self.session_mgr.commit()

    def finish_interrupted_request (self, request, exc):
        output = Publisher.finish_interrupted_request(self, request, exc)

        # commit the current transaction so that any changes to the
        # session objects are saved and are visible on the next HTTP
        # hit.  Remember, AccessError is a subclass of PublishError,
        # so this code will be run for both typos in the URL and for
        # the user not being logged in.
        #
        # The assumption here is that the UI code won't make changes
        # to the core database before checking permissions and raising
        # a PublishError; if you must do this (though it's hard to see
        # why this would be necessary), you'll have to abort the
        # current transaction, make your session changes, and then
        # raise the PublishError.
        # 
        # XXX We should really be able to commit session changes and
        # database changes separately, but that requires ZODB
        # incantations that we currently don't know.
        self.session_mgr.commit()

        return output

    def finish_failed_request (self, request):
        self.session_mgr.abort()
        return Publisher.finish_failed_request(self, request)

